/*
 * The Worst Balanced Posets
 *
 * cple.c
 * Counting Poset Linear Extensions
 */

/* Copyright 2002--2015, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

#ifdef _USE_POF2_TABLE_
extern SET           pof2_tbl[];
#endif

#ifdef _USE_MIN_ELEMENT_TABLE_
extern MIN_EL_TBL_EL min_element_tbl[];
#endif

void cple_module_init(void) {
}

void cple_module_clean(void) {
}

typedef struct {
  long ord; /* the DFS ordinal number */
  CPC  usc; /* the linear extensions count of the upper set */
  CPC  dsc; /* the linear extensions count of the down set */
} VERTEX;

/* WARNING: It is assumed that the lattice is implicitly initialized with zeros. */
static long ordinal_number = 0;
static VERTEX lattice[1 << N];

CPC lec(POSET const *p, CPC mt[N][N]) {
  CPC du, t;
  SET subset[N + 1], m_set[N], i_set[N], j_set;
  int i, j, k, m;

  /* initlialization, finally mt[i][j] = e(P+ij) */
  for (i = 0; i < N; i++)
    for (j = 0; j < N; j++)
      mt[i][j] = 0;

  /* from bottom to top DFS */
  ordinal_number++;
  k = 0;
  subset[0] = EMPTY_SET;
  m_set[0] = FULL_SET;
  lattice[EMPTY_SET].usc = 0;
  lattice[FULL_SET].usc = 1;
  lattice[FULL_SET].ord = ordinal_number;
  for (;;) {
    while (m_set[k]) {
      m = min_element(m_set[k]);
      if ((p->r[1][m] & ~subset[k]) == EMPTY_SET)
        break;
      m_set[k] ^= pof2(m);
    }
    if (m_set[k]) {
      m_set[k] ^= pof2(m);
      subset[k + 1] = subset[k] | pof2(m);
      if (lattice[subset[k + 1]].ord == ordinal_number)
        lattice[subset[k]].usc += lattice[subset[k + 1]].usc;
      else {
        k++;
        m_set[k] = FULL_SET ^ subset[k]; /* = full_set & ~subset[k]; */
        lattice[subset[k]].usc = 0;
      }
    }
    else if (k > 0) {
      lattice[subset[k]].ord = ordinal_number;
      lattice[subset[k - 1]].usc += lattice[subset[k]].usc;
      k--;
    }
    else
      break;
  }

  /* from top to bottom DFS */
  ordinal_number++;
  k = 0;
  subset[0] = m_set[0] = FULL_SET;
  i_set[0] = EMPTY_SET; /* i_set[0] is modified but never used. */
  lattice[FULL_SET].dsc = 0;
  lattice[EMPTY_SET].dsc = 1;
  lattice[EMPTY_SET].ord = ordinal_number;
  for (;;) {
    while (m_set[k]) {
      m = min_element(m_set[k]);
      if ((p->r[0][m] & subset[k]) == EMPTY_SET)
        break;
      m_set[k] ^= pof2(m);
    }
    if (m_set[k]) {
      m_set[k] ^= pof2(m);
      i_set[k] |= pof2(m);
      subset[k + 1] = subset[k] ^ pof2(m);
      if (lattice[subset[k + 1]].ord == ordinal_number)
        lattice[subset[k]].dsc += lattice[subset[k + 1]].dsc;
      else {
        k++;
        m_set[k] = subset[k];
        i_set[k] = EMPTY_SET;
        lattice[subset[k]].dsc = 0;
      }
    }
    else if (k > 0) {
      t = lattice[subset[k]].usc;
      while (i_set[k]) {
        i = min_element(i_set[k]);
        du = lattice[subset[k] ^ pof2(i)].dsc * t;
        /* j_set = full_set ^ subset[k];
           Pairs which are in relation do not need to be computed. */
        j_set = FULL_SET ^ (subset[k] | p->r[0][i]);
        while (j_set) {
          j = min_element(j_set);
          mt[i][j] += du;
          j_set ^= pof2(j);
        }
        i_set[k] ^= pof2(i);
      }
      lattice[subset[k]].ord = ordinal_number;
      lattice[subset[k - 1]].dsc += lattice[subset[k]].dsc;
      k--;
    }
    else
      break;
  }

  /* Pairs witch are in relation need only to be updated. */
  t = lattice[EMPTY_SET].usc;
  for (i = 0; i < N; i++) {
    j_set = p->r[0][i];
    while (j_set) {
      j = min_element(j_set);
      mt[i][j] = t;
      j_set ^= pof2(j);
    }
  }

  return t;
}
