/*
 * The Worst Balanced Posets
 *
 * main.c
 * Main module
 */

/* Copyright 2002--2015, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

static void main_init(void);
static void main_clean(void);

#ifdef _USE_POF2_TABLE_
extern SET           pof2_tbl[];
#endif
#ifdef _USE_MIN_ELEMENT_TABLE_
extern MIN_EL_TBL_EL min_element_tbl[];
#endif
#ifdef _USE_CARDINALITY_TABLE_
extern CARD_TBL_EL   cardinality_tbl[];
#endif

/*
 * Global data initialization and deinitialization
 */

static void main_init(void) {
  wbp_module_init();
  cple_module_init();
}

static void main_clean(void) {
  cple_module_clean();
  wbp_module_clean();
}

/*
 * Program administration
 */

void fatal_error(const char *fmt, ...) {
  va_list fmt_args;

  fprintf(stderr, "ERROR: ");
  va_start(fmt_args, fmt);
  vfprintf(stderr, fmt, fmt_args);
  va_end(fmt_args);
  if (errno) {
    fprintf(stderr, ", errno=%d", errno);
    perror(" ");
  }
  else
    fprintf(stderr, "\n");
  main_clean();
  exit(2);
}

int main(int argc, char *argv[]) {
  int result = 0;

  if (argc < 2) {
    printf("usage: %s action [parameters]\ncompiled at "__DATE__
           " "__TIME__"\n", argv[0]);
    #ifdef _NRCT_
    printf("non-recursive congruence testing\n");
    #else
    printf("recursive congruence testing\n");
    #endif
    #ifdef _INT_INDEX_
    printf("int indices\n");
    #else
    printf("long indices\n");
    #endif
    #if defined _USE_MIN_ELEMENT_TABLE_
      printf("min element table\n");
    #elif defined _USE_BUILT_IN_MIN_ELEMENT_
      printf("built-in min element\n");
    #endif
    #if defined _USE_BUILT_IN_MAX_ELEMENT_
      printf("built-in max element\n");
    #endif
    printf("N = %d   sizeof(POSET) = %d\n"
           "MAX_STACK = %ld   HASH_TBL_SIZE = %ld\n",
           N, (int)sizeof(POSET),
           (long)MAX_STACK, (long)HASH_TBL_SIZE);
    return 3;
  }
  switch (atoi(argv[1])) {
    case 15: /* the worst balanced poset */
      /* check_type: 0 - not a linear sum
                     1 - ladders with broken rungs
                     2 - not a linear sum, width at most 2
                     3 - not a linear sum, width at most 3
      */
      if (argc < 3) {
        printf("usage: %s 15 check_type\n\n", argv[0]);
        return 3;
      }
      printf("%s %s %s\n", argv[0], argv[1], argv[2]);
      main_init();
      result = wbp_iterative_search(atoi(argv[2]));
      break;

    default:
      printf("unknown action: %s\n", argv[1]);
      return 3;
  }
  main_clean();
  return result;
}

/*
 * Auxiliary routines
 */

void poset_print(const POSET *g) {
  int i, j, k;
  SET V;
  static const char description[2][6] = {"above", "below"};

  /* printf("n=%d e(P)=" CPC_PRINT_FORMAT "\n", g->n, g->t); */
  printf("n=%d\n", g->n);
  for (k = 0; k < 2; k++) {
    printf("%s\n", description[k]);
    for (i = 0; i < g->n; i++) {
      V = g->r[k][i];
      if (V) {
        printf("%d:", i);
        do {
          j = min_element(V);
          V ^= pof2(j);
          printf(" %d", j);
        } while(V);
        printf("\n");
      }
    }
  }
}
