/*
 * The Worst Balanced Posets
 *
 * match.c
 * Graph isomorphism routines - Ullman's algorithm
 */

/* Copyright 2002--2015, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "main.h"

#ifdef _USE_POF2_TABLE_
extern SET           pof2_tbl[];
#endif
#ifdef _USE_MIN_ELEMENT_TABLE_
extern MIN_EL_TBL_EL min_element_tbl[];
#endif
#ifdef _USE_CARDINALITY_TABLE_
extern CARD_TBL_EL   cardinality_tbl[];
#endif

#ifdef _NRCT_

static BOOL nr_match(SET const PMT[], SET const ra_1[], SET const ra_2[], int n) {
  int  i, j, k, l[N]; /* k is perm_len, l[k] is node2 */
  BOOL possible;
  SET  MT[N][N], j_set;

  k = 0;
  for (i = 0; i < n; i++) {
    l[i] = 0;
    MT[0][i] = PMT[i];
  }
  for (;;) {
    possible = FALSE;
    if (in(MT[k][k], l[k])) {
      possible = TRUE;
      for (i = k + 1; i < n; i++) {
        j_set = MT[k + 1][i] = MT[k][i]; /* & ~pof2(l[k]); */
        /* If there exists only one of two edges:
           (i,k) in graph 1 or (j,l) in graph 2
           or exists only one of two edges:
           (k,i) in graph 1 or (l,j) in graph 2
           then isomorphism can't map vertex i to vertex j. */
        while (j_set) {
          j = min_element(j_set);
          if (in(ra_1[i], k) != in(ra_2[j], l[k]) ||
              in(ra_1[k], i) != in(ra_2[l[k]], j))
            MT[k + 1][i] ^= pof2(j);
          j_set ^= pof2(j);
        }
        MT[k + 1][i] &= ~pof2(l[k]);
        if (MT[k + 1][i] == EMPTY_SET) {
          possible = FALSE;
          break;
        }
      }
    }
    if (possible) {
      if (k == n - 2)
        return TRUE;
      else
        k++;
    }
    else {
      while (k >= 0 && l[k] >= n - 1) {
        l[k] = 0;
        k--;
      }
      if (k >= 0)
        l[k]++;
      else
        return FALSE;
    }
  }
}

BOOL congruent(POSET const *g1, POSET const *g2) {
  int c10[N], c11[N], c20[N], c21[N];
  int i, j, n;
  SET MT[N];

  n = g1->n;
  for (i = 0; i < n; i++) {
    c10[i] = cardinality(g1->r[0][i]);
    c20[i] = cardinality(g2->r[0][i]);
    c11[i] = cardinality(g1->r[1][i]);
    c21[i] = cardinality(g2->r[1][i]);
  }
  for (i = 0; i < n; i++) {
    MT[i] = EMPTY_SET;
    for (j = 0; j < n; j++)
      if (c10[i] == c20[j] && c11[i] == c21[j])
        MT[i] |= pof2(j);
    if (MT[i] == EMPTY_SET)
      goto TEST_DUALISM;
  }
  if (nr_match(MT, g1->r[0], g2->r[0], n))
    return TRUE;
TEST_DUALISM:
  for (i = 0; i < n; i++) {
    MT[i] = EMPTY_SET;
    for (j = 0; j < n; j++)
      if (c11[i] == c20[j] && c10[i] == c21[j])
        MT[i] |= pof2(j);
    if (MT[i] == EMPTY_SET)
      return FALSE;
  }
  return nr_match(MT, g1->r[1], g2->r[0], n);
}

#else

typedef struct {
  int n;           /* the number of vertices */
  int perm_len;    /* the number of matched vertices */
  int perm[N];     /* found permutation - isomorphism */
  SET MT[N];       /* matrix encoding the compatibility of vertices */
  SET const *ra_1; /* compared graph */
  SET const *ra_2; /* compared graph */
} ULL_STATE;

static BOOL ull_match_1(ULL_STATE *p_s) {
  ULL_STATE s1;
  int       i, j, node2;
  BOOL      found;
  SET       nodes, j_set;

  /* Find vertex node2 in graph 2 which can match vertex perm_len in graph 1. */
  found = FALSE;
  nodes = p_s->MT[p_s->perm_len];
  while (!found && nodes != EMPTY_SET) {
    node2 = min_element(nodes);
    nodes ^= pof2(node2);

    /* Test if matching found. */
    if (p_s->perm_len == p_s->n - 1) {
      /* There was the last par verticies to check, isomorphism found. */
      return TRUE;
    }
    else {
      /* Try to extend permutation (isomorphism) by one vertex. */
      s1 = *p_s;
      s1.perm[p_s->perm_len] = node2;
      s1.perm_len++;

      /* Remove all pairs from the matrix M which are not compatible
         with isomorphism conditions. */
      for (i = s1.perm_len; i < s1.n; i++) {
        s1.MT[i] &= ~pof2(node2);
        /* for (j = 0; j < s1.n; j++) if (s1.MT[i] & pof2(j)) ... */
        j_set = s1.MT[i];
        while (j_set) {
          j = min_element(j_set);
          /* If there exists only one of two edges:
             (i,k) in graph 1 or (j,l) in graph 2
             or exists only one of two edges:
             (k,i) in graph 1 or (l,j) in graph 2. */
          if (in(s1.ra_1[i], p_s->perm_len) != in(s1.ra_2[j], node2) ||
              in(s1.ra_1[p_s->perm_len], i) != in(s1.ra_2[node2], j)) {
            /* Isomorphism can't map vertex i to vertex j. */
            /* s1.MT[i] &= ~pof2(j); */
            s1.MT[i] ^= pof2(j);
          }
          j_set ^= pof2(j);
        }
        /* Test if still exists possibility to match. */
        if (s1.MT[i] == EMPTY_SET) goto TEST_NEXT_NODE;
      }

      /* Maping vertex perm_len in graph 1 to vertex node2 in graph 2
         is possible, call self recursively. */
      found = ull_match_1(&s1);
    }
TEST_NEXT_NODE:
    ;
  }
  return found;
}

BOOL congruent(const POSET *g1, const POSET *g2) {
  ULL_STATE s0;
  int       c10[N], c11[N], c20[N], c21[N], i, j;

  s0.n = g1->n;
  for (i = 0; i < s0.n; i++) {
    c10[i] = cardinality(g1->r[0][i]);
    c20[i] = cardinality(g2->r[0][i]);
    c11[i] = cardinality(g1->r[1][i]);
    c21[i] = cardinality(g2->r[1][i]);
  }
  for (i = 0; i < s0.n; i++) {
    s0.MT[i] = EMPTY_SET;
    for (j = 0; j < s0.n; j++)
      if (c10[i] == c20[j] && c11[i] == c21[j])
        s0.MT[i] |= pof2(j);
    if (s0.MT[i] == EMPTY_SET)
      goto TEST_DUALISM;
  }
  s0.perm_len = 0;
  s0.ra_1     = g1->r[0];
  s0.ra_2     = g2->r[0];
  if (ull_match_1(&s0))
    return TRUE;
TEST_DUALISM:
  for (i = 0; i < s0.n; i++) {
    s0.MT[i] = EMPTY_SET;
    for (j = 0; j < s0.n; j++)
      if (c11[i] == c20[j] && c10[i] == c21[j])
        s0.MT[i] |= pof2(j);
    if (s0.MT[i] == EMPTY_SET)
      return FALSE;
  }
  s0.perm_len = 0;
  s0.ra_1     = g1->r[1];
  s0.ra_2     = g2->r[0];
  return ull_match_1(&s0);
}

#endif
